#include <Xmd/Buffer.h>
#include <string.h>
#include <stdio.h>

#define XmdBUFFER_GROWTH_FACTOR    2
#define XmdBUFFER_INITIAL_CAPACITY 8

struct _XmdBuffer {
	void *   data;
	Cardinal size;
	Cardinal length;
	Cardinal capacity;
};

static void  XmdBufferResize (XmdBuffer *buffer, Cardinal length);
static void  XmdBufferFit    (XmdBuffer *buffer);
static void *XmdBufferOffset (XmdBuffer *buffer, Cardinal index);

XmdBuffer *XmdBufferNew (Cardinal size) {
	XmdBuffer *buffer = XtNew(XmdBuffer);
	buffer->size     = size;
	buffer->length   = 0;
	buffer->capacity = XmdBUFFER_INITIAL_CAPACITY;
	buffer->data     = XtCalloc(buffer->capacity, buffer->size);
	XmdBufferFit(buffer);
	return buffer;
}

void XmdBufferResize (XmdBuffer *buffer, Cardinal length) {
	buffer->length = length;
	XmdBufferFit(buffer);
}

void XmdBufferFit (XmdBuffer *buffer) {
	if (buffer->length > buffer->capacity) {
		buffer->capacity *= XmdBUFFER_GROWTH_FACTOR;
		buffer->data = XtReallocArray (
			buffer->data,
			buffer->capacity,
			buffer->size);
	}
}

void *XmdBufferOffset (XmdBuffer *buffer, Cardinal index) {
	return buffer->data + (index * buffer->size);
}

void XmdBufferPush (XmdBuffer *buffer, void *element) {
	XmdBufferResize(buffer, buffer->length + 1);
	memcpy (
		XmdBufferOffset(buffer, buffer->length - 1),
		element,
		buffer->size);
}

void *XmdBufferBreak (XmdBuffer *buffer) {
	void *data = buffer->data;
	XtFree((char *)(buffer));
	return data;
}

void XmdBufferFree (XmdBuffer *buffer) {
	XtFree(buffer->data);
	XtFree((char *)(buffer));
}
